import random
import pyglet
import const


class GameWindow(pyglet.window.Window):
    
    def __init__(self):
        super(GameWindow, self).__init__(const.WIDTH, const.HEIGHT, const.TITLE)
        #self.start_intro()
        self.start_game()
    
    def start_game(self):
        self.current_window = Game(self)
        try:
            player = pyglet.media.Player()
            player.queue(pyglet.resource.media('ironhorse.ogg', streaming=True))
            player.eos_action = player.EOS_LOOP
            player.play()
        except:
            print 'Unable to load music.'
    
    def start_intro(self):
        self.current_window = Intro(self)


class Intro:
    
    def __init__(self, window):
        self.window = window
        self.batch = pyglet.graphics.Batch()
        
        self.window.push_handlers(self.on_draw)
    
    def on_draw(self):
        self.window.clear()
        self.batch.draw()


class WorldMap:
    
    def __init__(self, window):
        self.window = window
        self.batch = pyglet.graphics.Batch()
        
        self.window.push_handlers(self.on_draw)
    
    def on_draw(self):
        self.window.clear()
        self.batch.draw()


class Game:

    def __init__(self, window):
        self.window = window
        self.batch = pyglet.graphics.Batch()
        self.bg_batch = pyglet.graphics.Batch()
        self.floor_batch = pyglet.graphics.Batch()
        self.target_batch = pyglet.graphics.Batch()
        self.cage_batch = pyglet.graphics.Batch()
        self.score_batch = pyglet.graphics.Batch()
        
        bg_image = pyglet.resource.image('bg.png')
        self.bg = pyglet.sprite.Sprite(bg_image, 0, 0, batch=self.bg_batch)
        
        self.crane = Crane(self.batch, self.cage_batch, (const.WIDTH/2, const.HEIGHT-80))
        self.score = Score(self.score_batch)
        self.floor_mgr = FloorManager(self.floor_batch, self.target_batch, self.score, self.crane)
        self.collision_mgr = CollisionManager(self.crane, self.floor_mgr, self.score)

        self.paused = False
        self.paused_label = None

        self.window.push_handlers(self.on_draw, self.on_key_press, self.on_key_release)
        pyglet.clock.schedule_interval(self.update, const.INV_FPS)

    def finish(self):
        self.window.pop_handlers()
        pyglet.clock.unschedule(self.update)

    def on_draw(self):
        self.window.clear()
        self.bg_batch.draw()
        self.floor_batch.draw()
        self.target_batch.draw()
        self.cage_batch.draw()
        self.batch.draw()
        self.score_batch.draw()
    
    def pause(self):
        self.paused = True
        self.paused_label = pyglet.text.Label('Paused', batch=self.batch,
            x=const.WIDTH / 2, y=const.HEIGHT / 2, font_size=36,
            anchor_x='center', anchor_y='center', color=(255,255,255,255))

    def unpause(self):
        self.paused = False
        self.paused_label.delete()
        self.paused_label = None
    
    def on_key_press(self, symbol, modifiers):
        if symbol == pyglet.window.key.P:
            if not self.paused: self.pause()
            else: self.unpause()
        elif symbol == pyglet.window.key.LEFT or symbol == pyglet.window.key.A:
            self.crane.start_move_left()
        elif symbol == pyglet.window.key.RIGHT or symbol == pyglet.window.key.D:
            self.crane.start_move_right()
        elif symbol in [pyglet.window.key.DOWN, pyglet.window.key.Z, pyglet.window.key.X,
                        pyglet.window.key.SPACE]:
            self.crane.start_dropping()
    
    def on_key_release(self, symbol, modifiers):
        if symbol == pyglet.window.key.LEFT or symbol == pyglet.window.key.A:
            self.crane.stop_move_left()
        elif symbol == pyglet.window.key.RIGHT or symbol == pyglet.window.key.D:
            self.crane.stop_move_right()
        elif symbol in [pyglet.window.key.DOWN, pyglet.window.key.Z, pyglet.window.key.X,
                        pyglet.window.key.SPACE]:
            self.crane.stop_dropping()

    def update(self, dt):
        if self.paused: return
        if self.score.done:
            self.paused = True
            if self.score.win: text = 'You win!'
            else: text = 'Missed too many!'
            self.paused_label = pyglet.text.Label(text, batch=self.batch,
                x=const.WIDTH / 2, y=const.HEIGHT / 2, font_size=36,
                anchor_x='center', anchor_y='center', color=(255,255,255,255))
            return
        self.crane.update(dt)
        self.floor_mgr.update(dt)
        self.collision_mgr.update(dt)


class Target(pyglet.sprite.Sprite):
    
    def __init__(self, batch, floor, x, ttype=0, lazy=False):
        #image = pyglet.image.SolidColorImagePattern((0,0,50,255)).create_image(50, 50)
        img_full = pyglet.resource.image(const.TARGET_NAMES[ttype] + '.png')
        img_seq = pyglet.image.ImageGrid(img_full, 1, const.TARGET_TOTAL_FRAMES[ttype])
        if lazy:
            image = pyglet.image.Animation.from_image_sequence(img_seq[const.TARGET_DIV_FRAME[ttype]:], const.TARGET_RATE_LAZY[ttype], True)
        else:
            image = pyglet.image.Animation.from_image_sequence(img_seq[:const.TARGET_DIV_FRAME[ttype]], const.TARGET_RATE_NORMAL[ttype], True)
        super(Target, self).__init__(image, batch=batch)
        self.x = floor.x + x
        self.y = floor.y
        self.floor = floor
        self.vel = [0, 0]
        self.lazy = lazy
        self.ttype = ttype

    def update(self, dt):
        self.x += self.vel[0] * dt
        self.y = self.floor.y
        

class Cage(pyglet.sprite.Sprite):

    def __init__(self, batch, pos):
        image = pyglet.resource.image('cage.png')
        super(Cage, self).__init__(image, batch=batch)
        self.x = pos[0]
        self.y = pos[1]
        self.orig_y = pos[1]
        self.vel = [0, 0]
        self.accel = [0, 0]
        self.drop_speed = 600
        self.gravity = -1500
        self.max_speed = 1200
        self.dropped = False
        self.hit_floor = None
        self.done = False

    def update(self, dt):
        self.vel[0] += self.accel[0] * dt
        self.vel[1] += self.accel[1] * dt
        if self.vel[1] < -self.max_speed: self.vel[1] = -self.max_speed
        
        if not self.dropped:
            self.x += self.vel[0] * dt
        self.y += self.vel[1] * dt
        if self.y + self.height <= - 2 * self.height or self.y > const.HEIGHT:
            self.done = True
        
        if self.hit_floor:
            self.y = self.hit_floor.y
        
    def drop(self):
        self.vel[1] = -self.drop_speed
        self.accel[1] = self.gravity
        self.dropped = True

    def new_cage_drop(self):
        self.vel[1] = -self.drop_speed

    def stop_dropping(self):
        self.vel[1] = 0
    
    def reset(self):
        self.dropped = False
        self.y = self.orig_y
        self.vel[1] = 0
        self.accel[1] = 0


class Crane(pyglet.sprite.Sprite):

    def __init__(self, batch, cage_batch, pos, limit=const.WIDTH):
        image = pyglet.resource.image('crane.png')
        super(Crane, self).__init__(image, batch=batch)
        self.cage_batch = cage_batch
        self.x = pos[0]
        self.y = pos[1]
        self.vel = [0, 0]
        self.cage_offset_y = 20
        self.cages = [Cage(cage_batch, self.cage_pos())]
        self.current_cage = self.cages[0]
        self.speed = 400
        self.dropping = False
        self.set_limit(limit)
        self.dropping_new_cage = False
    
    def set_limit(self, limit):
        self.left_limit = const.WIDTH / 2 - limit / 2 - const.CRANE_ARM_WIDTH
        self.right_limit = const.WIDTH / 2 + limit / 2 + const.CRANE_ARM_WIDTH - 2

    def cage_pos(self):
        return [self.x + self.width / 2 - 32, self.y - self.cage_offset_y]

    def start_move_left(self):
        self.vel[0] = -self.speed

    def start_move_right(self):
        self.vel[0] = self.speed

    def stop_move_left(self):
        if self.vel[0] < 0:
            self.vel[0] = 0

    def stop_move_right(self):
        if self.vel[0] > 0:
            self.vel[0] = 0

    def start_dropping(self):
        self.dropping = True

    def stop_dropping(self):
        self.dropping = False

    def new_cage(self):
        if self.current_cage: return # only add cage if no cage
        cp = self.cage_pos()
        self.cages.append(Cage(self.cage_batch, (cp[0], const.HEIGHT)))
        self.current_cage = self.cages[-1]
        self.dropping_new_cage = True
        self.current_cage.new_cage_drop()

    def update(self, dt):
        if self.dropping and self.current_cage and not self.dropping_new_cage:
            self.current_cage.drop()
            self.current_cage = None
            self.new_cage()
        
        self.x += self.vel[0] * dt
        self.y += self.vel[1] * dt
        if self.x < self.left_limit: self.x = self.left_limit
        if self.x + self.width > self.right_limit: self.x = self.right_limit - self.width
        
        remove_list = []
        for cage in self.cages:
            cage.update(dt)
            if cage.done:
                remove_list.append(cage)
        for cage in remove_list:
            self.cages.remove(cage)
            cage.delete()

        if self.dropping_new_cage and self.current_cage.y < self.y - self.cage_offset_y:
            self.current_cage.y = self.y - self.cage_offset_y
            self.current_cage.stop_dropping()
            self.dropping_new_cage = False
        self.current_cage.x = self.cage_pos()[0]


class Floor(pyglet.sprite.Sprite):
    
    def __init__(self, batch, pos, width, speed=40):
        color_back = '%c%c%c%c' % (255, 255, 255, 255)
        color_wall = '%c%c%c%c' % (0, 0, 0, 255)
        data = ''
        for i in xrange(const.FLOOR_HEIGHT - const.WALL_THICKNESS):
            data += color_wall * const.WALL_THICKNESS + color_back * width + color_wall * const.WALL_THICKNESS
        data += color_wall * (width + 2 * const.WALL_THICKNESS) * const.WALL_THICKNESS
        image = pyglet.image.ImageData(width + 2 * const.WALL_THICKNESS, const.FLOOR_HEIGHT, 'RGBA', data)
        super(Floor, self).__init__(image, batch=batch)
        self.x = pos[0] - self.width / 2
        self.y = pos[1]
        self.vel = [0, speed]
        
        self.targets = []
        self.done = False
        self.past_cage_line = False

        self.earthquake_mode = False
        self.active = False
    
    def add_target(self, target):
        self.targets.append(target)

    def update(self, dt):
        self.x += self.vel[0] * dt
        self.y += self.vel[1] * dt
        if self.earthquake_mode:
            if self.x <= 0 or self.x + self.width >= const.WIDTH:
                self.vel[0] = -self.vel[0]
        
        for target in self.targets:
            target.update(dt)

        if self.y > const.HEIGHT:
            self.done = True
    
    def stop(self):
        self.vel[1] = 0
    
    def deactivate(self):
        self.set_color((64, 64, 64))
        self.active = False
        
    def deactivate_bad(self):
        self.set_color((255, 64, 64))
        self.active = False
        
    def deactivate_good(self):
        self.set_color((64, 255, 64))
        self.active = False

    def activate(self):
        self.set_color((255, 255, 255))
        self.active = True
        
    def set_color(self, color):
        self.color = color
        for target in self.targets: target.color = color
        
    def passed_cage_line(self):
        self.past_cage_line = True
        if self.active:
            self.deactivate_bad()

    def start_earthquake(self):
        self.vel[0] = const.EARTHQUAKE_SPEED * random.randint(-1, 1) # 0 included
        self.earthquake_mode = True


class Earth(pyglet.sprite.Sprite):

    def __init__(self, batch, pos_y, speed=40):
        image = pyglet.image.SolidColorImagePattern((20, 5, 5, 255)).create_image(const.WIDTH, const.HEIGHT)
        super(Earth, self).__init__(image, batch=batch)
        self.x = 0
        self.y = pos_y
        self.vel = [0, speed]

    def update(self, dt):
        self.y += self.vel[1] * dt

    def stop(self):
        self.vel[1] = 0


class FloorManager():
    
    def __init__(self, batch, target_batch, score, crane, width=310, num_floors=const.NUM_FLOORS, speed=60):
        self.num_floors = num_floors
        self.current_floor = self.num_floors
        self.batch = batch
        self.target_batch = target_batch
        self.score = score
        self.speed = speed
        self.crane = crane
        
        self.floors = []
        self.floor_x = const.WIDTH / 2
        self.active_floor = -1
        self.earth = None

        self.larger_floor(2)
        self.height = const.FLOOR_HEIGHT

        self.cage_line = const.HEIGHT - 100

        self.all_floors_done = False
        self.done = False
        self.earthquake_mode = False

        self.generate_random()
    
    def generate(self, targets_x, ttypes, lazy):
        if len(self.floors) == 0: cur_y = -self.height
        else: cur_y = self.floors[-1].y - self.height
        floor = Floor(self.batch, (self.floor_x, cur_y), self.width, self.speed)
        self.floors.append(floor)

        if self.earthquake_mode:
            floor.start_earthquake()
            floor.x = random.randint(0, const.WIDTH - self.width)

        for i in xrange(len(targets_x)):
            if i == lazy: lazy_set = True
            else: lazy_set = False
            floor.add_target(Target(self.target_batch, floor, targets_x[i], ttypes[i], lazy_set))

        floor.deactivate()
        
        if self.active_floor == -1:
            self.set_active_floor(0)
        elif self.active_floor == -2:
            self.set_active_floor(len(self.floors) - 1)
        self.current_floor -= 1
        self.check_next_level()
        if self.current_floor <= 0:
            self.all_floors_done = True
            self.earth = Earth(self.batch, self.floors[-1].y - const.HEIGHT, self.speed)

    def check_generate(self):
        if self.floors[-1].y >= -self.height and not self.all_floors_done:
            self.generate_random()

    def generate_random(self):
        targets_x = range(10, self.width - 10, 50)
        lazy = random.randint(0, len(targets_x) - 1)
        ttypes = random.choice(const.PATTERNS)
        self.generate(targets_x, ttypes, lazy)

    def next_floor(self):
        if self.active_floor + 1 == len(self.floors):
            self.active_floor = -2 # next floor
        else:
            self.set_active_floor(self.active_floor + 1)

    def set_active_floor(self, i):
        if i < len(self.floors):
            self.floors[self.active_floor].deactivate()
            self.floors[i].activate()
            self.active_floor = i
            self.crane.set_limit(self.floors[self.active_floor].width - 2 * const.WALL_THICKNESS)

    def get_active_floor(self):
        if self.active_floor >= 0:
            return self.floors[self.active_floor]
        return None
        
    def larger_floor(self, num_targets):
        self.num_targets = num_targets
        self.width = self.num_targets * 50 + 20
    
    def update(self, dt):
        self.check_generate()
        for floor in self.floors:
            floor.update(dt)
        if self.earth: self.earth.update(dt)
        
        # check all floors done
        if self.all_floors_done and self.floors[-1].y >= const.FLOOR_LINE:
            self.done = True
            for i in xrange(len(self.floors)):
                self.floors[i].stop()
                self.floors[i].y = const.FLOOR_LINE + (len(self.floors) - 1 - i) * self.height
                if self.earth: self.earth.stop()

        # assume only one floor done per frame
        floor = self.floors[0]
        if floor.done:
            self.floors.pop(0)
            if self.active_floor >= 0:
                self.active_floor -= 1
            if self.active_floor == 0:
                self.set_active_floor(0)
            floor.delete()

        # assume only the first two floors can be past cage line
        for i in xrange(min(2, len(self.floors))):
            floor = self.floors[i]
            if floor.y > self.cage_line and not floor.past_cage_line:
                floor.passed_cage_line()
                if self.active_floor == i:
                    self.next_floor()
                    self.score.miss()
                    self.floors[i].deactivate_bad()
    
    def check_next_level(self):
        # hardcoded, whee!
        if self.current_floor in [65, 55, 45, 35, 25, 15, 5]:
            self.speed += 5
            for floor in self.floors:
                floor.vel[1] += 5
        if self.current_floor in [60, 50, 40, 30, 20, 10]:
            self.larger_floor(self.num_targets + 2)


class CollisionManager():

    def __init__(self, crane, floor_mgr, score):
        self.crane = crane
        self.floor_mgr = floor_mgr
        self.score = score

    def update(self, dt):
        floor = self.floor_mgr.get_active_floor()
        if not floor: return
        for cage in self.crane.cages:
            if not cage.hit_floor and cage != self.crane.current_cage and cage.y < floor.y:
                self.floor_mgr.next_floor()
                # not optimized, but not many targets anyway
                target_hit = False
                for target in floor.targets:
                    if self.hit_target_cage(target, cage):
                        if target.lazy:
                            target_hit = True
                            self.score.hit()
                            floor.deactivate_good()
                if not target_hit:
                    self.score.miss()
                    floor.deactivate_bad()
                cage.hit_floor = floor
                cage.y = floor.y

    def hit_target_cage(self, target, cage):
        return (cage.x - const.CAGE_THRESHOLD <= target.x and
                cage.x + cage.width + const.CAGE_THRESHOLD >= target.x + target.width)

    def hit_sprite(self, sprite1, sprite2):
        s1x = sprite1.x - sprite1.image.anchor_x
        s2x = sprite2.x - sprite2.image.anchor_x
        s1y = sprite1.y - sprite1.image.anchor_y
        s2y = sprite2.y - sprite2.image.anchor_y
        return (not (s1x > s2x + sprite2.width or
                     s1x + sprite1.width < s2x or
                     s1y + sprite1.height < s2y or
                     s1y > s2y + sprite2.height))


class Level():

    def __init__(self):
        self.speed = 30
        self.num_floor = 10
        self.width = 300


class Score():

    def __init__(self, batch):
        self.batch = batch
        self.score = 0
        self.combo = 0 # number of hits in a row
        self.max_combo = 0
        self.combo_multiplier = 1
        self.misses = 0
        self.num_floors = 0
        self.done = False
        self.win = False
        self.init_text()

    def hit(self):
        self.score += self.combo_multiplier
        self.combo += 1
        self.num_floors += 1
        if self.combo > self.max_combo:
            self.max_combo = self.combo
        if self.combo_multiplier < 5 and self.combo >= const.COMBO_LEVELS[self.combo_multiplier]:
            self.set_combo_multiplier(self.combo_multiplier + 1)
        self.update_text()
        if self.num_floors >= const.NUM_FLOORS:
            self.done = self.win = True

    def miss(self):
        self.misses += 1
        self.combo = 0
        self.num_floors += 1
        self.set_combo_multiplier(1)
        self.update_text()
        if self.misses > const.LIVES:
            self.done = True
        elif self.num_floors >= const.NUM_FLOORS:
            self.done = self.win = True

    def set_combo_multiplier(self, mult):
        self.combo_multiplier = mult

    def init_text(self):
        # self.score_label = pyglet.text.Label('Score: ' + str(self.score),
        #     batch=self.batch, x=30, y=const.HEIGHT-30, font_size=12,
        #     anchor_x='left', anchor_y='top', color=(255,255,255,255))
        # self.combo_label = pyglet.text.Label('Combo: ' + str(self.combo),
        #     batch=self.batch, x=30, y=const.HEIGHT-60, font_size=12,
        #     anchor_x='left', anchor_y='top', color=(255,255,255,255))
        # self.multiplier_label = pyglet.text.Label('Multiplier: ' + str(self.combo_multiplier) + 'x',
        #     batch=self.batch, x=30, y=const.HEIGHT-90, font_size=12,
        #     anchor_x='left', anchor_y='top', color=(255,255,255,255))
        self.misses_label = pyglet.text.Label('Misses: ' + str(self.misses) + ' (' + str(const.LIVES) + ' allowed)',
            batch=self.batch, x=30, y=30, font_size=12,
            anchor_x='left', anchor_y='top', color=(128,128,255,255))
        self.nfloors_label = pyglet.text.Label('Number of floors left: ' + str(const.NUM_FLOORS - self.num_floors),
            batch=self.batch, x=const.WIDTH-30, y=30, font_size=12,
            anchor_x='right', anchor_y='top', color=(128,128,255,255))

    def update_text(self):
        # self.score_label.text = 'Score: ' + str(self.score)
        # self.combo_label.text = 'Combo: ' + str(self.combo)
        # self.multiplier_label.text = 'Multiplier: ' + str(self.combo_multiplier) + 'x'
        self.misses_label.text = 'Misses: ' + str(self.misses)  + ' (' + str(const.LIVES) + ' allowed)'
        self.nfloors_label.text = 'Number of floors left: ' + str(const.NUM_FLOORS - self.num_floors)